#ifndef CALOUT_H
#define CALOUT_H

#include "turbid.h"
#include "tprops.h"
#include <complex>
#include <alsa/asoundlib.h>

typedef complex<double> cmplx;

class rotor{
public:
  double checkphase;
// total phase is iphase + subphase
  double iphase;        // will always represent an integer (for thousands of years)
  double subphase;      // will always be between zero and 1
  double ph_step;
  cmplx acc;            // accumulator
  cmplx c_step;

// constructor
  rotor() :
    checkphase(0),
    iphase(0),
    subphase(0),
    acc(1, 0)
  {}
// this is the PREFIX increment operator:
  inline rotor& operator++(){
    subphase += ph_step;
    double ipart = floor(subphase);
    subphase -= ipart;
    iphase += ipart;
#if 1
// this uses 49% of a cpu, for one channel at 440 Hz
// i.e. 109 spikes
// ... or 32% of a cpu, for one channel at 1000 Hz
// i.e. 48 spikes
// ... so it is slightly more efficient
//
// With smaller block size, 46% at 1000 Hz.
// With smaller block size, 80% at  440 Hz ...
//   ... with occasional reader underruns.
    acc *= c_step;
    if (iphase >= checkphase) {
      cmplx reset(cos(twopi*subphase), sin(twopi*subphase));
      cmplx delta(acc - reset);
      if (abs(delta) > 1e-8) {
        cout << "huh? delta: " << delta
           << "  iphase: " << iphase
           << "  subphase: " << subphase
           << "  acc:  "  << acc
           << "  c_step: " << c_step
           << "  reset: " << reset
        << endl;
      }
      acc = reset;
      checkphase += 10000.;
    }
#else
// this uses 49% of a cpu, for one channel at 1000 Hz
// i.e. 48 spikes:
    acc = cmplx(1, sin(twopi*subphase));
#endif
    return *this;
  }
// phase measured in circular units (not radians)
  void init(double const dphase_dframe){
    ph_step = dphase_dframe;
    c_step = cmplx(cos(twopi*ph_step), sin(twopi*ph_step));
  }
};


class Keeper{
public:
  alsa_dsp* dsp;
  int const fpp;                // frames per period
  int nchan;
  std::vector<datum> buf;
  int finbuf;                   // frames in buffer at the moment
  Keeper(alsa_dsp*, int const fsize);
  snd_pcm_sframes_t writei(const datum* buffer,
        snd_pcm_uframes_t appbuf__fr);

};

typedef vector<rotor> chord;          // one rotor per spike

class calibrator{
public:
  tprops & tpx;
  alsa_dsp & cal_dsp;
  int oblock;
  datum* ptr;
  double sinamp, squamp, broad_factor, comb_amp;
  int broad_phase;
  vector<rotor> sine_rotor;
  double nyquist;
  snd_pcm_uframes_t calblock__frames;
  vector<datum> calout_block;
  unsigned int activeNsig;
  int comb_interval;
  int nspike;
  vector<chord> comb_rotor;
  Keeper keep;

// constructor
  calibrator(tprops& _tpx, alsa_dsp& _dsp)
  : tpx(_tpx),
    cal_dsp(_dsp),
    oblock(0),
    keep(&cal_dsp, FPP)
  {}

// other methods
  int setup();
  virtual void io_step();
  virtual vector<pollfd> pollist();
};

#endif /* CALOUT_H */
