// Double buffering rationale:
// 1) Small buffer would be bad because demand for randomness
// is often very bursty.  When a spike of demand comes along,
// you don't want to wait while entropy is harvested.
// 2) Small buffers cause unnecessarily many context switches,
// passing buffers through the kernel.
// 3) A large single buffer would be bad in case of something
// with a slow, steady demand -- lots of jitter when the
// generator needs to refill the buffer.
// 4) Double-buffering with good-sized buffers is just fine.
//
// Note in contrast linux /dev/random very commonly returns
// a byte at a time under conditions of high demand.  Ick.
//

using namespace std;

#include <vector>
#include <sys/types.h>		/* for fork() */
#include <unistd.h>		/* for fork */
#include <syslog.h>	/* needed although not mentioned by man page */
//#incl <linux/unistd.h> /* not needed although mentioned by man page */
#include "dbuffer.h"
#include "turbid.h"

int dbuffer(const int pipe[2], const int outfd, const int obs){

  int kidpid = fork();
  if (kidpid < 0) {
    syslog(LOG_ERR, "Fork failed (for dbuffer): %m");
    exeunt(1);
  }
  if (kidpid) {			// the parent
		// caller should close(pipe[0])
		// parent has no need to read pipe
		// ... and closing it makes broken-pipe stuff work
    return kidpid;
  }
// kid code follows.
// Note kid doesn't need to do any exeunts;
// the mainline code will notice a broken pipe
// if anything happens to the dbuffer kid process.
  progname = "turbid-dbuffer";
  close(pipe[1]);		// child has no need to write pipe
		// ... and closing it makes broken-pipe stuff work
  int infd(pipe[0]);

  vector<char> buf(obs);
  char* ptr(&buf[0]);
  for (;;) {
    int ngot = read(infd, ptr, obs);
    if (ngot == 0) exit(0);		// EoF
    if (ngot < 0) {
      syslog(LOG_ERR, "dbuffer read error: %m");
      exit(1);
    }
    int rslt = write(outfd, ptr, ngot);
    if (rslt != ngot) {
      syslog(LOG_ERR, "dbuffer write error (%d): %m", rslt);
      if (rslt > 0) {} 		/* short write, maybe OK */
      else exit(1);
    }
  }
}
