// Double buffering rationale:
// 1) Small buffer would be bad because demand for randomness
// is often very bursty.  When a spike of demand comes along,
// you don't want to wait while entropy is harvested.
// 2) Small buffers cause unnecessarily many context switches,
// passing buffers through the kernel.
// 3) A large single buffer would be bad in case of something
// with a slow, steady demand -- lots of jitter when the
// generator needs to refill the buffer.
// 4) Double-buffering with good-sized buffers is just fine.
//
// Note in contrast linux /dev/random very commonly returns
// a byte at a time under conditions of high demand.  Ick.
//
// productivity = 3.6 megabytes per second on a 1GHz pentium
// productivity = 10.4 megabytes per second
//   on a "Core 2 Duo"  @ 2.80 MHZ
//   cpu MHz 800
//   bogomips 5586.48

using namespace std;

#include <stdint.h>
#include <iostream>
#include <iomanip>              /* for hex */
#include <sys/types.h>		/* for fork() */
#include <unistd.h>		/* for fork */
#include <syslog.h>	/* needed although not mentioned by man page */
//#include <linux/unistd.h>	/* not needed although mentioned */
#include <assert.h>
#include "stretch.h"
#include "turbid.h"
#include "sha1.h"
#include "sha1-more.h"
#include "bufwrite.h"

// parent returns immediately
// return-value is kidpid, so caller can keep track of what kid is doing

int stretch(const int pipe[2], const int stretchfd, const int obs){
  int kidpid = fork();
  if (kidpid < 0) {
    syslog(LOG_ERR, "Fork failed (for stretch): %m");
    exeunt(1);
  }
  if (kidpid) {			// the parent
		// caller should close(pipe[0])
		// parent has no need to read pipe
		// ... and closing it makes broken-pipe stuff work
    return kidpid;
  }
// kid code follows.
// Note kid doesn't need to do any exeunts;
// the mainline code will notice a broken pipe
// if anything happens to the stretch kid process.
  progname = "turbid-stretch";
  close(pipe[1]);		// child has no need to write pipe
		// ... and closing it makes broken-pipe stuff work
  int const & infd(pipe[0]);

// end of process-control stuff
//////////////////////////////////////////
// begin actual randomizing

// The stretch factor is the producte of inner_stretch*outer_stretch.
// It's roughly inversely proportional to the entropy density of the output.
  int inner_stretch(10);
  int outer_stretch(1000);
  SHA1_CTX main_ctx;		// SHA-1 "context"
  SHA1_CTX inner_ctx;
  Digest main_digest;
  Digest inner_digest;
  int main_counter(0);

// buffered output channel:
  bufwrite chunky(stretchfd, obs);
  for (;;) {                    // over all seeds

// start by reading a seed:
    int want_read = sizeof(inner_digest);
    unsigned char* ptr(inner_digest.byte);
    while (want_read){
      int did_read = read(infd, ptr, want_read);
      ///????? cerr << ".";
      if (did_read == 0) {
        syslog(LOG_ERR, "stretch read EoF");
        exit(1);
      }
      if (did_read < 0) {
        syslog(LOG_ERR, "stretch read error: %m");
        exit(1);
      }
      ptr += did_read;
      want_read -= did_read;
    }

    for (int jj = 0; jj < outer_stretch; jj++) {          // 1000 times
      if (jj) {
  // Irreversible sequencer.

#if 0
Elaine Barker and John Kelsey
NIST SP 800-90A,
"Recommendation for Random Number Generation Using Deterministic Random Bit Generators"
(January 2012)
http://csrc.nist.gov/publications/nistpubs/800-90A/SP800-90A.pdf


Section 8.8 i.e. page 24 discusses "backtracking resistance"
Section 10.2 i.e. page 49 covers "DRBG Mechanisms Based on Block Ciphers"

#endif

  // Needed to produce Forward Secrecy.
  // A regular counter is far too reversible.
  // A LFSR is far too reversible.
  // This is slightly expensive, so we only do it 1 time in 10
  // The output of this sequencer MUST NOT be exposed to the outside.
        SHA1_Init(&inner_ctx);
        SHA1_Update(&inner_ctx, inner_digest.byte, sizeof(inner_digest));
        SHA1_Final(inner_digest.byte, &inner_ctx);
      }
      for (int ii = 0; ii < inner_stretch; ii++){         // 10 times
        main_counter++;                // don't care if it wraps around
        SHA1_Init(&main_ctx);
  // The reversible counter:
        SHA1_Update(&main_ctx, (unsigned char*) &main_counter,
                                  sizeof(main_counter));
  // The irreversible sequencer:
        SHA1_Update(&main_ctx, inner_digest.byte, sizeof(inner_digest));
  // Some feedback, just for fun:
  // Probably not helpful, since it was exposed to the outside world:
        SHA1_Update(&main_ctx, main_digest.byte, sizeof(main_digest));
        SHA1_Final(main_digest.byte, &main_ctx);
  // This write will block if there is no need for our services:
        chunky.Write(main_digest.byte, sizeof(main_digest));
        /// ???? cerr << "%";
      } // inner stretch
    }   // outer stretch
  }     // end loop over all seeds
}
